/* vim: cin ts=8 sw=4 syn=c
 *
 * $Id$
 *
 * Copyright (c) 2007 Garry T. Williams */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <poll.h>
#include <errno.h>
#include <unistd.h>
#include "watch_dir.h"

#define BUFSIZE 8192

int main(int argc, char *argv[]) {
    int     in_fd;
    int     watch_descriptor;
    char   *buf;
    struct  pollfd pollfd[1];
    size_t  buf_size = BUFSIZE;
    int     ev_cnt = 0;

    if (argc != 2) {
	fprintf(stderr, "usage: %s watch-dir\n", argv[0]);
	exit(1);
    }

    if ((in_fd = inotify_init()) < 0) {
	perror("inotify_init");
	exit(1);
    }

    if ((watch_descriptor
		= inotify_add_watch(in_fd, argv[1], IN_ALL_EVENTS)) < 0)
    {
	perror("inotify_add_watch");
	exit(1);
    }

    {
	struct sigaction act;

	act.sa_handler = catch_sig;
	act.sa_flags   = 0;
	(void)sigemptyset(&act.sa_mask);

	if (sigaction(SIGINT, &act, NULL) < 0) {
	    perror("sigaction");
	    exit(1);
	}

	if (sigaction(SIGTERM, &act, NULL) < 0) {
	    perror("sigaction");
	    exit(1);
	}

	if (sigaction(SIGQUIT, &act, NULL) < 0) {
	    perror("sigaction");
	    exit(1);
	}
    }

    if ((buf = malloc(buf_size)) == NULL) {
	perror("malloc");
	exit(1);
    }

    pollfd->fd     = in_fd;
    pollfd->events = POLLIN;

    while (1) {
	size_t  len;
	char   *eob;
	char   *events;
	int     cnt = 0;

	/* wait for event(s) */
	if (poll(pollfd, 1, -1) < 0) {
	    if (errno == EINTR)
		break;

	    perror("poll");
	    exit(1);
	}

	/* This assumes that only whole inotify_event records are ever
	 * ready to read */
	if (ioctl(in_fd, FIONREAD, &len) < 0) {
	    perror("ioctl");
	    exit(1);
	}

	if (len > buf_size) {
	    if ((buf = realloc(buf, len)) == NULL) {
		perror("realloc");
		exit(1);
	    }

	    buf_size = len;
	}

	{
	    size_t bytes_read = read(in_fd, buf, len);

	    if (bytes_read < 0) {
		if (errno == EINTR)
		    break;

		perror("read");
		exit(1);
	    }

	    if (!bytes_read)
		break;

	    eob = buf + bytes_read;
	}

	ev_cnt++;

	/* process the event(s) */
	events = buf;
	while (events < eob) {
	    struct inotify_event *event = (struct inotify_event *)events;

	    if (watch_descriptor != event->wd)
		printf("got WD %i with 0x%X; expected WD %i\n",
			event->wd, event->mask, watch_descriptor);

	    else
		handle_event(event, ev_cnt, ++cnt);

	    events += sizeof(struct inotify_event) + event->len;
	}
    }

    (void)inotify_rm_watch(in_fd, watch_descriptor);
    free(buf);
    exit(0);
}

static void handle_event(struct inotify_event *event, int ev_cnt, int cnt) {
    char *file = "(no file name)";

    if (event->len)
	file = event->name;

    if (event->mask & IN_ACCESS)
	printf("(%i,%i) ACCESS: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_ATTRIB)
	printf("(%i,%i) METADATA: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_CLOSE_WRITE)
	printf("(%i,%i) CLOSED OUTPUT: %s mask: 0x%X\n", ev_cnt, cnt,
		file, event->mask);

    if (event->mask & IN_CLOSE_NOWRITE)
	printf("(%i,%i) CLOSED INPUT: %s mask: 0x%X\n", ev_cnt, cnt,
		file, event->mask);

    if (event->mask & IN_CREATE)
	printf("(%i,%i) CREATED: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_DELETE)
	printf("(%i,%i) DELETED: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_DELETE_SELF) {
	printf("(%i,%i) WATCH DIR DELETED: %s mask: 0x%X\n", ev_cnt,
		cnt, file, event->mask);
	exit(0);
    }

    if (event->mask & IN_MODIFY)
	printf("(%i,%i) MODIFIED: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_MOVE_SELF) {
	printf("(%i,%i) MOVED: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);
	exit(0);
    }

    if (event->mask & IN_MOVED_FROM)
	printf("(%i,%i) MOVED OUT: %s cookie: 0x%X mask: 0x%X\n",
		ev_cnt, cnt, file, event->cookie, event->mask);

    if (event->mask & IN_MOVED_TO)
	printf("(%i,%i) MOVED IN: %s cookie: 0x%X mask: 0x%X\n",
		ev_cnt, cnt, file, event->cookie, event->mask);

    if (event->mask & IN_OPEN)
	printf("(%i,%i) OPENED: %s mask: 0x%X\n", ev_cnt, cnt, file,
		event->mask);

    if (event->mask & IN_IGNORED) {
	printf("(%i,%i) WATCH REMOVED: %s mask: 0x%X\n", ev_cnt, cnt,
		file, event->mask);
	exit(0);
    }
}

/* This is used only to get the poll() or read() to end with EINTR
 * when we receive a signal.  This requires POSIX semantics for
 * signals. */
static void catch_sig(int signo) {
    return;
}

